@@ -10,7 +10,7 @@
  *
  * USAGE:
  *
- * newrole [ -r role ] [ -t type ] [ -l level ] [ -- args ]
+ * newrole [ -r role ] [ -t type ] [ -l level ] [ -V ] [ -- args ]
  *
  * BUILD OPTIONS:
  *
@@ -62,6 +62,16 @@
 #include <selinux/get_default_type.h>
 #include <selinux/get_context_list.h> /* for SELINUX_DEFAULTUSER */
 #include <signal.h>
+#ifdef USE_AUDIT
+#include <libaudit.h>
+#endif
+#ifdef LOG_AUDIT_PRIV
+#ifndef USE_AUDIT
+#error LOG_AUDIT_PRIV needs the USE_AUDIT option
+#endif
+#include <sys/prctl.h>
+#include <sys/capability.h>
+#endif
 #ifdef USE_NLS
 #include <locale.h>			    /* for setlocale() */
 #include <libintl.h>			    /* for gettext() */
@@ -74,7 +84,7 @@
 #endif
 
 /* USAGE_STRING describes the command-line args of this program. */
-#define USAGE_STRING "USAGE: newrole [ -r role ] [ -t type ] [ -l level ] [ -- args ]"
+#define USAGE_STRING "USAGE: newrole [ -r role ] [ -t type ] [ -l level ] [ -V ] [ -- args ]"
 
 #define DEFAULT_CONTEXT_SIZE 255  /* first guess at context size */
 
@@ -93,7 +103,8 @@ char *xstrdup(const char *s)
 static char *
 build_new_range(char *newlevel, const char *range)
 {
-  char *newrangep = NULL, *tmpptr;
+  char *newrangep = NULL;
+  const char *tmpptr;
   size_t len;
 
   /* a missing or empty string */
@@ -147,7 +158,7 @@ build_new_range(char *newlevel, const char *range)
 
 #define SERVICE_NAME "newrole"    /* the name of this program for PAM */
 
-int authenticate_via_pam( const struct passwd * );
+int authenticate_via_pam( const struct passwd *, const char * );
 
 /* authenticate_via_pam()
  *
@@ -165,10 +176,12 @@ int authenticate_via_pam( const struct passwd * );
  *
  */
 
-int authenticate_via_pam( const struct passwd *pw ) {
+int authenticate_via_pam( const struct passwd *pw, const char *ttyn ) {
 
   int result = 0;    /* our result, set to 0 (not authenticated) by default */
+  int rc;	     /* pam return code */
   pam_handle_t *pam_handle;      /* opaque handle used by all PAM functions */
+  const char *tty_name;
 
   /* This is a jump table of functions for PAM to use when it wants to *
    * communicate with the user.  We'll be using misc_conv(), which is  *
@@ -180,26 +193,41 @@ int authenticate_via_pam( const struct passwd *pw ) {
 
   /* Make `p_pam_handle' a valid PAM handle so we can use it when *
    * calling PAM functions.                                       */
-  if( PAM_SUCCESS != pam_start( SERVICE_NAME,
-				pw->pw_name,
-				&pam_conversation,
-				&pam_handle ) ) {
+  rc = pam_start( SERVICE_NAME,
+		pw->pw_name,
+		&pam_conversation,
+		&pam_handle );
+  if( rc != PAM_SUCCESS ) {
     fprintf( stderr, _("failed to initialize PAM\n") );
     exit( -1 );
   }
 
+  if( strncmp(ttyn, "/dev/", 5) == 0 )
+    tty_name = ttyn+5;
+  else
+    tty_name = ttyn;
+
+  rc = pam_set_item( pam_handle, PAM_TTY, tty_name );
+  if( rc != PAM_SUCCESS ) {
+    fprintf( stderr, _("failed to set PAM_TTY\n") );
+    goto out;
+  }
+
   /* Ask PAM to authenticate the user running this program */
-  if( PAM_SUCCESS == pam_authenticate(pam_handle,0) ) {
-    result = 1;  /* user authenticated OK! */
+  rc = pam_authenticate(pam_handle,0);
+  if( rc != PAM_SUCCESS ) {
+    goto out;
   }
 
   /* Ask PAM to verify acct_mgmt */
-  if( PAM_SUCCESS != pam_acct_mgmt(pam_handle,0) ) {
-    result = 0;  
-  }
+  rc = pam_acct_mgmt(pam_handle,0);
+  if( rc == PAM_SUCCESS ) {
+    result = 1;  /* user authenticated OK! */
+  } 
 
   /* We're done with PAM.  Free `pam_handle'. */
-  pam_end( pam_handle, PAM_SUCCESS );
+out:
+  pam_end( pam_handle, rc );
  
   return( result );
 
@@ -277,6 +305,92 @@ int authenticate_via_shadow_passwd( const struct passwd *pw ) {
 
 #endif /* if/else USE_PAM */
 
+/*
+ * This function checks to see if the shell is known in /etc/shells.
+ * If so, it returns 1. On error or illegal shell, it returns 0.
+ */
+static int verify_shell(const char *shell_name)
+{
+  int found = 0;
+  const char *buf;
+
+  if( !shell_name )
+    return found;
+
+  while( (buf = getusershell()) != NULL ) {
+    /* ignore comments */
+    if( *buf == '#' )
+      continue;
+
+    /* check the shell skipping newline char */
+    if( ! strcmp (shell_name, buf) ) {
+      found = 1;
+      break;
+    }
+  }
+  endusershell();
+  return found;
+}
+
+/*
+ * This function will drop the capabilities so that we are left
+ * only with access to the audit system. If the user is root, we leave
+ * the capabilities alone since they already should have access to the
+ * audit netlink socket.
+ */
+#ifdef LOG_AUDIT_PRIV
+static void drop_capabilities(void)
+{
+  uid_t uid = getuid();
+
+  if (uid) { /* Non-root path */
+    cap_t new_caps, tmp_caps;
+    cap_value_t cap_list[] = { CAP_AUDIT_WRITE };
+    cap_value_t tmp_cap_list[] = { CAP_AUDIT_WRITE, CAP_SETUID };
+
+    new_caps = cap_init();
+    tmp_caps = cap_init();
+    if (!new_caps || !tmp_caps) {
+      fprintf(stderr, _("Error initing capabilities, aborting.\n"));
+      exit(-1);
+    }
+    cap_set_flag(new_caps, CAP_PERMITTED, 1, cap_list, CAP_SET);
+    cap_set_flag(new_caps, CAP_EFFECTIVE, 1, cap_list, CAP_SET);
+    cap_set_flag(tmp_caps, CAP_PERMITTED, 2, tmp_cap_list, CAP_SET);
+    cap_set_flag(tmp_caps, CAP_EFFECTIVE, 2, tmp_cap_list, CAP_SET);
+
+    /* Keep capabilities across uid change */
+    prctl(PR_SET_KEEPCAPS, 1, 0, 0, 0);
+
+    /* We should still have root's caps, so drop most capabilities now */
+    if (cap_set_proc(tmp_caps)) {
+      fprintf(stderr, _("Error dropping capabilities, aborting\n"));
+      exit(-1);
+    }
+    cap_free(tmp_caps);
+
+    /* Change uid */
+    if (setresuid(uid, uid, uid)) {
+      fprintf(stderr, _("Error changing uid, aborting.\n"));
+      exit(-1);
+    }
+
+    /* Now get rid of this ability */
+    if (prctl(PR_SET_KEEPCAPS, 0, 0, 0, 0) < 0) {
+      fprintf(stderr, _("Error resetting KEEPCAPS, aborting\n"));
+      exit(-1);
+    }
+
+    /* Finish dropping capabilities. */
+    if (cap_set_proc(new_caps)) {
+      fprintf(stderr, _("Error dropping SETUID capability, aborting\n"));
+      exit(-1);
+    }
+    cap_free(new_caps);
+  }
+}
+#endif
+
 /************************************************************************
  *
  * All code used for both PAM and shadow passwd goes in this section.
@@ -285,25 +399,25 @@ int authenticate_via_shadow_passwd( const struct passwd *pw ) {
 
 int main( int argc, char *argv[] ) {
 
-  security_context_t new_context=NULL;          /* our target security context */
+  security_context_t new_context=NULL;	   /* our target security context */
   security_context_t old_context=NULL;	   /* our original securiy context */
   security_context_t tty_context=NULL;	   /* The current context of tty file */
-  security_context_t new_tty_context=NULL;	   /* The new context of tty file */
+  security_context_t new_tty_context=NULL; /* The new context of tty file */
   security_context_t chk_tty_context= NULL;
 
-  context_t context;		 	   /* manipulatable form of new_context */
+  context_t context;		 	 /* manipulatable form of new_context */
 
 
-  const char *se_username;  /* SELinux user identity */
   struct passwd *pw;                 /* struct derived from passwd file line */
   struct passwd pw_copy;
 
   int clflag;                        /* holds codes for command line flags */
   int flag_index;                    /* flag index in argv[] */
-  struct option long_options[] = {   /* long option flags for getopt() */
+  const struct option long_options[] = {   /* long option flags for getopt() */
     { "role", 1, 0, 'r' },
     { "type", 1, 0, 't' },
     { "level", 1, 0, 'l' },
+    { "version", 0, 0, 'V' },
     { NULL, 0, 0, 0 }
   };
   char *role_s = NULL;               /* role spec'd by user in argv[] */
@@ -311,7 +425,18 @@ int main( int argc, char *argv[] ) {
   char *level_s = NULL;              /* level spec'd by user in argv[] */
   char *ttyn   = NULL;		     /* tty path */
   pid_t childPid=0;			     
-  int fd=0;
+  uid_t uid;
+  int fd;
+  int enforcing;
+  sigset_t empty;
+
+#ifdef LOG_AUDIT_PRIV
+  drop_capabilities();
+#endif
+
+  /* Empty the signal mask in case someone is blocking a signal */
+  sigemptyset( &empty );
+  (void) sigprocmask( SIG_SETMASK, &empty, NULL );
 
   /* Terminate on SIGHUP. */
   signal(SIGHUP, SIG_DFL);
@@ -334,13 +459,22 @@ int main( int argc, char *argv[] ) {
 	     _("Sorry, newrole may be used only on a SELinux kernel.\n") );
     exit(-1);
   }
+  enforcing = security_getenforce();
+  if (enforcing < 0) {
+    fprintf(stderr, _("Could not determine enforcing mode.\n"));
+    exit(-1);
+  }
 
   while (1) {
-    clflag=getopt_long(argc,argv,"r:t:l:",long_options,&flag_index);
+    clflag=getopt_long(argc,argv,"r:t:l:V",long_options,&flag_index);
     if (clflag == -1)
 	break;
     
     switch( clflag ) {
+    case 'V':
+	printf("newrole: %s version %s\n",PACKAGE,VERSION);
+	exit(0);
+	break;
     case 'r':
       /* If role_s is already set, the user spec'd multiple roles - bad. */
       if( role_s ) {
@@ -430,18 +564,24 @@ int main( int argc, char *argv[] ) {
     exit(-1);
   }
 
-  freecon(old_context);
-  /* Make `pw' point to a structure containing the data              *
-   * from our user's line in the passwd file.  If the current user's
-   * SELinux user identity is the default (SELINUX_DEFAULTUSER), then
-   * we authenticate using the user's UID.  Otherwise we use the SELinux
-   * user identity.
+  /*
+   * Determine the Linux user identity to re-authenticate.
+   * If supported and set, use the login uid, as this should be more stable.
+   * Otherwise, use the real uid.
+   * The SELinux user identity is no longer used, as Linux users are now
+   * mapped to SELinux users via seusers and the SELinux user identity space
+   * is separate.
    */
-  se_username = context_user_get(context);
-  if (!strcmp (se_username, SELINUX_DEFAULTUSER))
-    pw = getpwuid(getuid());
-  else
-    pw=getpwnam(se_username);
+#ifdef USE_AUDIT
+  uid = audit_getloginuid();
+  if (uid == (uid_t)-1)
+	  uid = getuid();
+#else
+  uid = getuid();
+#endif    
+  
+  /* Get the passwd info for the Linux user identity. */
+  pw = getpwuid(uid);
   if( !pw ) {
     fprintf(stderr,_("cannot find your entry in the passwd file.\n"));
     exit(-1);
@@ -452,11 +592,28 @@ int main( int argc, char *argv[] ) {
   pw->pw_dir = xstrdup(pw->pw_dir);
   pw->pw_shell = xstrdup(pw->pw_shell);
 
+  if (verify_shell(pw->pw_shell) == 0) {
+	fprintf(stderr, _("Error!  Shell is not valid.\n"));
+	exit(-1);
+  }
+
+  /* Get the tty name. Pam will need it. */
+  ttyn=ttyname(0);
+  if (!ttyn || *ttyn == '\0') {
+	fprintf(stderr, _("Error!  Could not retrieve tty information.\n"));
+	exit(-1);
+  }
+
   printf(_("Authenticating %s.\n"),pw->pw_name);
 
-  /* Authenticate the user running this program. */
+  /* 
+   * Re-authenticate the user running this program.
+   * This is just to help confirm user intent (vs. invocation by
+   * malicious software), not to authorize the operation (which is covered
+   * by policy).  Trusted path mechanism would be preferred.
+   */
 #ifdef USE_PAM
-  if( !authenticate_via_pam(pw) ) 
+  if( !authenticate_via_pam(pw, ttyn) ) 
 #else /* !USE_PAM */
   if( !authenticate_via_shadow_passwd(pw) ) 
 #endif /* if/else USE_PAM */
@@ -548,31 +705,30 @@ int main( int argc, char *argv[] ) {
    *
    */
 
-  /* Fetch TTY information */
-  ttyn=ttyname(0);
-  if (!ttyn || *ttyn == '\0') {
-	fprintf(stderr, _("Warning!  Could not retrieve tty information.\n"));
-	exit(-1);
-  }
-
+  /* Re-open TTY descriptor */
   fd = open(ttyn, O_RDWR);
   if (fd < 0) {
-	fprintf(stderr, _("Warning!  Could not open %s.\n"), ttyn);
+	fprintf(stderr, _("Error!  Could not open %s.\n"), ttyn);
 	exit(-1);
   }
 
   tty_context = NULL;
-  if (fgetfilecon(fd, &tty_context) < 0)
-	fprintf(stderr, _("Warning!  Could not get current context for %s, not relabeling.\n"), ttyn);
-    
+  if (fgetfilecon(fd, &tty_context) < 0) {
+	fprintf(stderr, _("%s!  Could not get current context for %s, not relabeling tty.\n"), enforcing ? "Error" : "Warning", ttyn);
+        if (enforcing)
+	  exit(-1);
+  }
 #ifdef CANTSPELLGDB
   if (tty_context)
     printf("Your tty %s was labeled with context %s\n", ttyn, tty_context);
 #endif
 
   new_tty_context = NULL;
-  if (tty_context && (security_compute_relabel(new_context,tty_context,SECCLASS_CHR_FILE,&new_tty_context) < 0))
-       fprintf(stderr, _("Warning!  Could not get new context for %s, not relabeling.\n"), ttyn);
+  if (tty_context && (security_compute_relabel(new_context,tty_context,SECCLASS_CHR_FILE,&new_tty_context) < 0)) {
+       fprintf(stderr, _("%s!  Could not get new context for %s, not relabeling tty.\n"), enforcing ? "Error" : "Warning", ttyn);
+       if (enforcing)
+         exit(-1);
+  }
 
 #ifdef CANTSPELLGDB
   if (new_tty_context)
@@ -581,9 +737,11 @@ int main( int argc, char *argv[] ) {
 
   if (new_tty_context) {
     if (fsetfilecon(fd,new_tty_context) < 0) {
-      fprintf(stderr, _("Warning!  Could not set new context for %s\n"), ttyn);
+      fprintf(stderr, _("%s!  Could not set new context for %s\n"), enforcing ? "Error" : "Warning", ttyn);
       freecon(new_tty_context);
       new_tty_context = NULL;
+      if (enforcing)
+        exit(-1);
     }
   }
 
@@ -592,12 +750,18 @@ int main( int argc, char *argv[] ) {
   if( childPid<0 ) {
     int errsv=errno;
     fprintf(stderr,_("newrole: failure forking: %s"),strerror(errsv));
+    if (fsetfilecon(fd,tty_context) < 0)
+      fprintf(stderr, _("Warning!  Could not restore context for %s\n"), ttyn);
+    freecon(tty_context);
     exit(-1);
   } else if (childPid) {
     /* PARENT */
-    wait(NULL);
+    int rc;
+    do {
+      rc = wait(NULL);
+    } while (rc < 0 && errno == EINTR);
 
-    if (!new_tty_context)
+    if( !new_tty_context || !tty_context )
       exit(0);
 
     /* Verify that the tty still has the context set by newrole. */
@@ -634,15 +798,15 @@ int main( int argc, char *argv[] ) {
       fprintf(stderr,_("Could not close descriptors.\n"));
       exit(-1);
     }
-  fd = open(ttyn,O_RDWR);
+  fd = open(ttyn,O_RDONLY);
   if (fd != 0) {
       exit(-1);
   }
-  fd = open(ttyn,O_RDWR);  
+  fd = open(ttyn,O_WRONLY);  
   if (fd != 1) {
       exit(-1);
   }
-  fd = open(ttyn,O_RDWR);  
+  fd = open(ttyn,O_WRONLY);  
   if (fd != 2) {
       exit(-1);
   }
@@ -668,6 +832,32 @@ int main( int argc, char *argv[] ) {
        fprintf(stderr, _("Could not set exec context to %s.\n"), new_context);
        exit(-1);
   }
+#ifdef LOG_AUDIT_PRIV
+  /* Send audit message */
+  {
+    char *msg;
+    int rc;
+    int audit_fd = audit_open();
+    if (audit_fd < 0) {
+       fprintf(stderr, _("Error connecting to audit system.\n"));
+       exit(-1);
+    }
+    if (asprintf(&msg, "newrole: old-context=%s new-context=%s",
+             old_context, new_context) < 0) {
+       fprintf(stderr, _("Error allocating memory.\n"));
+       exit(-1);
+    }
+    rc = audit_log_user_message(audit_fd, AUDIT_USER_ROLE_CHANGE,
+                               msg, NULL, NULL, ttyn, 1);
+    if (rc <= 0) {
+       fprintf(stderr, _("Error sending audit message.\n"));
+       exit(-1);
+    }
+    free(msg);
+    close(audit_fd);
+  }
+#endif
+  freecon(old_context);
   execv(argv[optind-1],argv+optind-1);
 
   /* If we reach here, then we failed to exec the new shell. */
