@@ -23,7 +23,9 @@
  *
  */
 
+#define _GNU_SOURCE
 #include <stdio.h>
+#include <stdio_ext.h>
 #include <errno.h>
 #include <string.h>
 #include <stdlib.h>
@@ -45,6 +47,8 @@ static char *progname;
 static int errors=0;
 static int recurse=0;
 static int force=0;
+#define STAT_BLOCK_SIZE 1
+static int pipe_fds[2] = { -1, -1 };
 
 #define MAX_EXCLUDES 100
 static int excludeCtr=0;
@@ -112,24 +116,16 @@ static int only_changed_user(const char *a, const char *b)
 void usage(const char * const name)
 {	
   fprintf(stderr,
-	  "usage:  %s [-Rnv] [-e excludedir ] [-o filename ] [-f filename | pathname... ]\n",  name);
+	  "usage:  %s [-FnrRv] [-e excludedir ] [-o filename ] [-f filename | pathname... ]\n",  name);
   exit(1);
 }
-int restore(char *filename) {
+/* filename has trailing '/' removed by nftw or other calling code */
+int restore(const char *filename) {
   int retcontext=0;
-  int retval=0;
   security_context_t scontext=NULL;
   security_context_t prev_context=NULL;
-  int len=strlen(filename);
   struct stat st;
   char path[PATH_MAX+1];
-  int user_only_changed=0;
-  /* 
-     Eliminate trailing /
-  */
-  if (len > 0 && filename[len-1]=='/' && (strcmp(filename,"/") != 0)) {
-    filename[len-1]=0;
-  }
 
   if (excludeCtr > 0 && exclude(filename)) {
       return 0;
@@ -143,9 +139,10 @@ int restore(char *filename) {
     if (verbose>1)
       fprintf(stderr,"Warning! %s refers to a symbolic link, not following last component.\n", filename);
     char *p = NULL, *file_sep;
-    char *tmp_path = strdup(filename);
+    char *tmp_path = strdupa(filename);
+    size_t len=0;
     if (!tmp_path) {
-      fprintf(stderr,"strdup on %s failed:  %s\n", filename,strerror(errno));
+      fprintf(stderr,"strdupa on %s failed:  %s\n", filename,strerror(errno));
       return 1;
     }
     file_sep = strrchr(tmp_path, '/');
@@ -154,15 +151,22 @@ int restore(char *filename) {
       *file_sep = 0;
       file_sep++;
       p = realpath(tmp_path, path);
+    } 
+    else {
+      file_sep = tmp_path;
+      p = realpath("./", path);
     }
-    if (!p || strlen(path) + strlen(file_sep) + 1 > PATH_MAX) {
+    if(p)
+      len = strlen(p);
+    if (!p || len + strlen(file_sep) + 2 > PATH_MAX) {
       fprintf(stderr,"realpath(%s) failed %s\n", filename, strerror(errno));
-      free(tmp_path);
       return 1;
     }
-    sprintf(p + strlen(p), "/%s", file_sep);
-    filename = p;
-    free(tmp_path);
+    p += len;
+    *p = '/';
+    p++;
+    strcpy(p, file_sep);
+    filename = path;
   } else {
     char *p;
     p = realpath(filename, path);
@@ -175,17 +179,12 @@ int restore(char *filename) {
   if (excludeCtr > 0 && exclude(filename)) {
       return 0;
   }
-  retval = matchpathcon(filename, st.st_mode, &scontext);
-  if (retval < 0) {
+  if (matchpathcon(filename, st.st_mode, &scontext) < 0) {
     if (errno == ENOENT)
       return 0;
     fprintf(stderr,"matchpathcon(%s) failed %s\n", filename,strerror(errno));
     return 1;
   } 
-  if (strcmp(scontext,"<<none>>")==0) {
-    freecon(scontext);
-    return 0;
-  }
   retcontext=lgetfilecon(filename,&prev_context);
   
   if (retcontext >= 0 || errno == ENODATA) {
@@ -194,27 +193,24 @@ int restore(char *filename) {
     if (retcontext < 0 || force || 
 	(strcmp(prev_context,scontext) != 0 && 	
 	 !(customizable=is_context_customizable(prev_context) > 0))) {
-      if (outfile) {
-	fprintf(outfile, "%s\n", filename);
-      }
-      user_only_changed = only_changed_user(scontext, prev_context);
-      if (change && !user_only_changed) {
-	retval=lsetfilecon(filename,scontext);
-      }
-      if (retval<0) {
-	  fprintf(stderr,"%s set context %s->%s failed:'%s'\n",
-		  progname, filename, scontext, strerror(errno));
-	  if (retcontext >= 0)
-	    freecon(prev_context);
-	  freecon(scontext);
-	  return 1;
-      } else 	
-	      if (verbose && 
-		  (verbose > 1 || !user_only_changed))
+      if (only_changed_user(scontext, prev_context) == 0) {
+	      if (outfile) fprintf(outfile, "%s\n", filename);
+	      if (change) {
+		      if (lsetfilecon(filename,scontext) < 0) {
+			      fprintf(stderr,"%s set context %s->%s failed:'%s'\n",
+				      progname, filename, scontext, strerror(errno));
+			      if (retcontext >= 0)
+				      freecon(prev_context);
+			      freecon(scontext);
+			      return 1;
+		      }
+	      }
+	      if (verbose)
 		      printf("%s reset %s context %s->%s\n",
-			      progname, filename, (retcontext >= 0 ? prev_context : ""), scontext);
+			     progname, filename, (retcontext >= 0 ? prev_context : ""), scontext);
+      }
     }
-    if (verbose > 1 && customizable>0) {
+    if (verbose > 1 && ! force && customizable>0) {
 	    printf("%s: %s not reset customized by admin to %s\n",
 		      progname, filename, prev_context);
     }
@@ -230,38 +226,80 @@ int restore(char *filename) {
   freecon(scontext);
   return errors;
 }
+
+static int pre_stat(const char *file_unused __attribute__((unused)),
+		      const struct stat *sb_unused __attribute__((unused)),
+		      int flag_unused __attribute__((unused)),
+		      struct FTW *s_unused __attribute__((unused)))
+{
+  char buf[STAT_BLOCK_SIZE];
+  if(write(pipe_fds[1], buf, STAT_BLOCK_SIZE) != STAT_BLOCK_SIZE)
+  {
+    fprintf(stderr, "Error writing to stat pipe, child exiting.\n");
+    exit(1);
+  }
+  return 0;
+}
+
 static int apply_spec(const char *file,
 		      const struct stat *sb_unused __attribute__((unused)),
 		      int flag,
 		      struct FTW *s_unused __attribute__((unused)))
 {
+	char buf[STAT_BLOCK_SIZE];
+	if(pipe_fds[0] != -1 && read(pipe_fds[0], buf, STAT_BLOCK_SIZE) != STAT_BLOCK_SIZE)
+	{
+		fprintf(stderr, "Read error on pipe.\n");
+		pipe_fds[0] = -1;
+	}
 	if (flag == FTW_DNR) {
 		fprintf(stderr, "%s:  unable to read directory %s\n",
 			progname, file);
 		return 0;
 	}
-	errors=errors+restore((char *)file);
+	errors=errors+restore(file);
 	return 0;
 }
 void process(char *buf) {
+      int rc;
       if (recurse) {
-	if (nftw
-	    (buf, apply_spec, 1024, FTW_PHYS)) {
-	  fprintf(stderr,
-		  "%s:  error while labeling files under %s\n",
+	if(pipe(pipe_fds) == -1)
+	  rc = -1;
+	else
+	  rc = fork();
+	if(rc == 0)
+	{
+	  close(pipe_fds[0]);
+	  nftw(buf, pre_stat, 1024, FTW_PHYS);
+	  exit(1);
+	}
+	if(rc > 0)
+	  close(pipe_fds[1]);
+	if(rc == -1 || rc > 0) {
+	  if (nftw(buf, apply_spec, 1024, FTW_PHYS)) {
+	    fprintf(stderr, "%s:  error while labeling files under %s\n",
 		  progname, buf);
-	  errors++;
+	    errors++;
+	  }
 	}
       }
       else
-	errors=errors+restore(buf);
+      {
+	/* Eliminate trailing / */
+	size_t len = strlen(buf);
+	if (len > 1 && buf[len - 1] == '/') {
+	  buf[len - 1] = 0;
+	}
+	errors = errors + restore(buf);
+      }
 }
 int main(int argc, char **argv) {
   int i=0;
   char *file_name=NULL;
   int file=0;
   int opt;
-  char buf[PATH_MAX];
+  char *buf = NULL;
+  size_t buf_len;
 
   memset(excludeArray,0, sizeof(excludeArray));
 
@@ -269,13 +307,12 @@ int main(int argc, char **argv) {
   if (is_selinux_enabled() <= 0 )
     exit(0);
 
-  memset(buf,0, sizeof(buf));
-
-  while ((opt = getopt(argc, argv, "FRnvf:o:e:")) > 0) {
+  while ((opt = getopt(argc, argv, "FrRnvf:o:e:")) > 0) {
     switch (opt) {
     case 'n':
       change = 0;
       break;
+    case 'r':
     case 'R':
       recurse = 1;
       break;
@@ -292,6 +329,7 @@ int main(int argc, char **argv) {
 		 optarg, strerror(errno));
 	 usage(argv[0]);
       }
+      __fsetlocking(outfile, FSETLOCKING_BYCALLER);
       break;
     case 'v':
       verbose++;
@@ -306,15 +344,16 @@ int main(int argc, char **argv) {
   }
   if (file) {
     FILE *f=stdin;
+    ssize_t len;
     if (strcmp(file_name,"-")!=0) 
       f=fopen(file_name,"r");
-	
     if (f==NULL) {
       fprintf(stderr,"Unable to open %s: %s\n", file_name, strerror(errno));
       usage(argv[0]);
     }
-    while(fgets(buf,PATH_MAX,f)) {
-      buf[strlen(buf)-1]=0;
+    __fsetlocking(f, FSETLOCKING_BYCALLER);
+    while((len = getline(&buf, &buf_len, f)) != -1) {
+      buf[len - 1] = 0;
       process(buf);
     }
     if (strcmp(file_name,"-")!=0) 
