@@ -41,6 +41,7 @@ static int num_commands = 0;
 static int verbose;
 static int reload;
 static int no_reload;
+static int create_store;
 static int build;
 
 static semanage_handle_t *sh = NULL;
@@ -169,9 +170,10 @@ static void parse_command_line(int argc, char **argv) {
         verbose = 0;
 	reload = 0;
 	no_reload = 0;
+	create_store = 0;
         while ((i = getopt_long(argc, argv, "s:b:hi:lvqr:u:RnB", opts, NULL)) != -1) {
                 switch (i) {
-                case 'b': set_mode(BASE_M, optarg); break;
+                case 'b': set_mode(BASE_M, optarg); create_store = 1; break;
                 case 'h': usage(argv[0]); exit(0);
                 case 'i': set_mode(INSTALL_M, optarg); break;
                 case 'l': set_mode(LIST_M, NULL); break;
@@ -222,7 +224,7 @@ int main(int argc, char *argv[]) {
         
 	sh = semanage_handle_create();
 	if (!sh) {
-                fprintf(stderr, "Could not create semanage handle\n");
+                fprintf(stderr, "%s:  Could not create semanage handle\n", argv[0]);
                 goto cleanup;
 	}
 
@@ -233,22 +235,37 @@ int main(int argc, char *argv[]) {
 		 * location */
 		semanage_select_store(sh, store, SEMANAGE_CON_DIRECT);
 	}
+	
+	/* if installing base module create store if necessary, for bootstrapping */
+	semanage_set_create_store(sh, create_store);
+	
+	if (!create_store) {
+		if (!semanage_is_managed(sh)) {
+			fprintf(stderr, "%s: SELinux policy is not managed or store cannot be accessed.\n", argv[0]);
+			goto cleanup;
+		}
+	
+		if (semanage_access_check(sh) < SEMANAGE_CAN_READ) {
+			fprintf(stderr, "%s: Cannot read policy store.\n", argv[0]);
+			goto cleanup;
+		}
+	}
 
         if ((result = semanage_connect(sh)) < 0) {
-                fprintf(stderr, "Could not connect to policy handler\n");
+                fprintf(stderr, "%s:  Could not connect to policy handler\n", argv[0]);
                 goto cleanup;
         }
 
 	if (reload) {
 		if ((result = semanage_reload_policy(sh)) < 0) {
-			fprintf(stderr, "Could not reload policy\n");
+			fprintf(stderr, "%s:  Could not reload policy\n", argv[0]);
 			goto cleanup;
 		}
 	}
 
 	if (build) {
 		if ((result = semanage_begin_transaction(sh)) < 0) {
-			fprintf(stderr, "Could not begin transaction\n");
+			fprintf(stderr, "%s:  Could not begin transaction\n", argv[0]);
 			goto cleanup;
 		}
 	}
@@ -260,7 +277,7 @@ int main(int argc, char *argv[]) {
                 size_t data_len;
                 if (mode == INSTALL_M || mode == UPGRADE_M || mode == BASE_M) {
                         if ((data_len = map_file(mode_arg, &data)) == 0) {
-                                fprintf(stderr, "Could not read file '%s':\n", mode_arg);
+                                fprintf(stderr, "%s:  Could not read file '%s':\n", argv[0], mode_arg);
                                 goto cleanup;
                         }
                 }
@@ -316,7 +333,8 @@ int main(int argc, char *argv[]) {
                         break;
                 }
                 default: {
-                        fprintf(stderr, "Unknown mode specified.\n");
+                        fprintf(stderr, "%s:  Unknown mode specified.\n", argv[0]);
+			usage(argv[0]);
                         goto cleanup;
                 }
                 }
@@ -325,7 +343,7 @@ int main(int argc, char *argv[]) {
                         munmap(data, data_len);
                 }
                 if (result < 0) {
-                        fprintf(stderr, "Failed!\n");
+			fprintf(stderr, "%s:  Failed on %s!\n", argv[0], mode_arg ?: "list");
                         goto cleanup;
                 }
                 else if (verbose) {
@@ -334,17 +352,17 @@ int main(int argc, char *argv[]) {
         }
         
         if (commit) {
-                if (verbose) {
-                        printf("Committing changes:\n");
-                }
-		if (no_reload) {
+		if (verbose)
+			printf("Committing changes:\n");
+		if (no_reload)
 			semanage_set_reload(sh, 0);
-		}
-                result = semanage_commit(sh);
+		if (build)
+			semanage_set_rebuild(sh, 1);
+ 		result = semanage_commit(sh);
         }
         
         if (result < 0) {
-                fprintf(stderr, "Failed!\n");
+                fprintf(stderr, "%s:  Failed!\n", argv[0]);
                 goto cleanup;
         }
         else if (commit && verbose) {
@@ -352,15 +370,17 @@ int main(int argc, char *argv[]) {
         }
 
         if (semanage_disconnect(sh) < 0) {
-                fprintf(stderr, "Error disconnecting\n");
+                fprintf(stderr, "%s:  Error disconnecting\n", argv[0]);
                 goto cleanup;
         }
         status = EXIT_SUCCESS;
 
  cleanup:
-        if (semanage_disconnect(sh) < 0) {
-		fprintf(stderr, "Error disconnecting\n");
-        }
+	if (semanage_is_connected(sh)) {
+	        if (semanage_disconnect(sh) < 0) {
+			fprintf(stderr, "%s:  Error disconnecting\n", argv[0]);
+		}
+	}
         semanage_handle_destroy(sh);
         cleanup();
         exit(status);
