@@ -160,9 +160,8 @@ void printf_tab(const char *outp) {
 
 int main(int argc, char **argv) {
 	/* these vars are reused several times */
-	int rc, i,c;
+	int rc,opt, i,c;
 	char *context;
-	char opt;
 
 	/* files that need context checks */
 	char *fc[MAX_CHECK];
